import firebase from 'firebase/app';

class EventEmitter {
    constructor() {
        this._listeners = [];
    }
    addEventListener(listener) {
        if (!this._listeners.includes(listener)) {
            this._listeners.push(listener);
        }
    }
    removeEventListener(listener) {
        const idx = this._listeners.indexOf(listener);
        if (idx !== -1) {
            this._listeners.splice(idx, 1);
        }
    }
    notify(uuid, event) {
        this._listeners.forEach((_listeners) => {
            _listeners(uuid, event);
        });
    }
}

function uuidv4() {
    return 'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx'.replace(/[xy]/g, function (c) {
        let r = (Math.random() * 16) | 0, v = c == 'x' ? r : (r & 0x3) | 0x8;
        return v.toString(16);
    });
}

class PeerRecord {
    constructor(username, uuid, ready) {
        this.username = username;
        this.uuid = uuid;
        this.ready = ready;
    }
}
class RoomRecord {
    constructor(roomId, peers) {
        this.roomId = roomId;
        this.peers = peers;
    }
}
class PairingRecord {
    constructor(status, uuid1, uuid2, iceCandidates1, iceCandidates2, offer, answer) {
        this.status = status;
        this.uuid1 = uuid1;
        this.uuid2 = uuid2;
        this.iceCandidates1 = iceCandidates1;
        this.iceCandidates2 = iceCandidates2;
        this.offer = offer;
        this.answer = answer;
    }
}
class PairingItem {
    constructor(iceCandidatesDone, answered, answerVerified) {
        this.iceCandidatesDone = iceCandidatesDone;
        this.answered = answered;
        this.answerVerified = answerVerified;
    }
}

class BaseSignaling {
    constructor() {
        this._uuid = '';
        this.roomRecord = null;
        this.pairingMap = new Map();
        this._roomRecordEmitter = new EventEmitter();
    }
    get roomRecordEmitter() {
        return this._roomRecordEmitter;
    }
    get uuid() {
        return this._uuid;
    }
    hostRoom(roomId, username, uuid) {
        if (!uuid)
            uuid = uuidv4();
        this._uuid = uuid;
        return this.internalHostRoom(roomId, username);
    }
    joinRoom(roomId, username, uuid) {
        if (!uuid)
            uuid = uuidv4();
        this._uuid = uuid;
        return this.internalJoinRoom(roomId, username);
    }
    upatePlayerStatus(ready) {
        if (this.roomRecord) {
            const peer = this.roomRecord.peers.find((peer) => peer.uuid === this._uuid);
            if (peer) {
                peer.ready = ready;
                this.saveRoomInfo();
            }
        }
    }
    roomRecordChanged(info, roomId, username, uuid) {
        if (!info || !info.peers) {
            this.roomRecord = new RoomRecord(roomId, []);
        }
        else {
            this.roomRecord = info;
        }
        if (!this.roomRecord.peers.some((peer) => peer.uuid === uuid)) {
            this.roomRecord.peers.push(new PeerRecord(username, uuid, false));
            this.saveRoomInfo();
        }
        else {
            this._roomRecordEmitter.notify(uuid, info);
        }
    }
    startPairings(mesh) {
        let pairing = true;
        return new Promise((resolve) => {
            mesh.meshReadyEmitter.addEventListener((_uuid, ready) => {
                this.cleanup();
                resolve(ready);
            });
            mesh.connectionReadyEmitter.addEventListener((_uuid, _ready) => {
                if (!pairing) {
                    return;
                }
                const connectionCount = mesh.connectionCount();
                const connectionsOpened = mesh.connectionsOpened();
                if (connectionsOpened === connectionCount) {
                    pairing = false;
                }
            });
            mesh.iceCandidateEmitter.addEventListener((uuid, candidate) => {
                let sCandidate = JSON.stringify(candidate);
                if (!candidate.candidate)
                    return;
                this.saveIceCandidate(uuid, sCandidate);
            });
            if (this.roomRecord) {
                const myIndex = this.roomRecord.peers.findIndex((peer) => peer.uuid === this._uuid);
                for (let [index, peer] of this.roomRecord.peers.entries()) {
                    if (index === myIndex) {
                        continue;
                    }
                    this.savePairing(mesh, myIndex, index, peer);
                }
            }
        });
    }
    pairingRecordChanged(mesh, info, peer, myIndex, index) {
        if (!info || !info.status) {
            if (myIndex < index) {
                mesh.createConnection(peer.uuid);
                setTimeout(() => {
                    mesh.createOffer(peer.uuid).then((offer) => {
                        info = new PairingRecord(1, this._uuid, peer.uuid, [], [], JSON.stringify(offer), '');
                        this.pairingMap.set(peer.uuid, new PairingItem([], false, false));
                        this.savePairingRecord(info, peer.uuid);
                    });
                }, 0);
            }
        }
        else {
            let pairingItem = this.pairingMap.get(peer.uuid);
            if (!pairingItem) {
                pairingItem = new PairingItem([], false, false);
                this.pairingMap.set(peer.uuid, pairingItem);
            }
            if (myIndex > index && info.offer && !pairingItem.answered) {
                pairingItem.answered = true;
                mesh.createConnection(peer.uuid);
                mesh
                    .createAnswer(peer.uuid, JSON.parse(info.offer))
                    .then((answer) => {
                    info.answer = JSON.stringify(answer);
                    this.savePairingRecord(info, peer.uuid);
                });
            }
            if (myIndex < index &&
                info.answer &&
                !pairingItem.answerVerified) {
                pairingItem.answerVerified = true;
                mesh.verifyAnswer(peer.uuid, JSON.parse(info.answer));
            }
            let candidates;
            if (myIndex < index) {
                candidates = info.iceCandidates2;
            }
            else {
                candidates = info.iceCandidates1;
            }
            candidates.forEach((candidate) => {
                if (pairingItem &&
                    !pairingItem.iceCandidatesDone.includes(candidate)) {
                    pairingItem.iceCandidatesDone.push(candidate);
                    const iceCandidate = JSON.parse(candidate);
                    if (iceCandidate.candidate) {
                        mesh.addIceCandidate(peer.uuid, iceCandidate);
                    }
                }
            });
        }
    }
}

class FirebaseSignaling extends BaseSignaling {
    constructor(config) {
        super();
        this.roomRef = null;
        this.pairRecordMap = new Map();
        firebase.initializeApp(config);
        this.db = firebase.firestore();
        this.roomsCollectionRef = this.db.collection('rooms');
    }
    cleanup() {
        // HACK: remove timer when where are sure all peers are connected
        // and can safely delete the info
        setTimeout(() => {
            if (this.roomRef)
                this.roomRef.delete();
            for (const value of this.pairRecordMap.values()) {
                value.delete();
            }
        }, 5000);
    }
    bindRoom(roomId, username) {
        return new Promise((resolve) => {
            this.roomRef = this.roomsCollectionRef.doc(roomId);
            this.roomRef.onSnapshot((doc) => {
                const data = doc.data();
                if (data) {
                    this.roomRecordChanged(JSON.parse(data.info), roomId, username, this._uuid);
                    resolve(true);
                }
                else {
                    resolve(false);
                }
            });
        });
    }
    internalHostRoom(roomId, username) {
        const roomRecord = new RoomRecord(roomId, []);
        this.roomRef = this.roomsCollectionRef.doc(roomId);
        return this.roomRef.set({ info: JSON.stringify(roomRecord) })
            .then(() => {
            return this.bindRoom(roomId, username);
        })
            .catch((error) => {
            console.log(error);
            return false;
        });
    }
    internalJoinRoom(roomId, username) {
        return this.bindRoom(roomId, username);
    }
    saveRoomInfo() {
        if (this.roomRef)
            this.roomRef.set({ info: JSON.stringify(this.roomRecord) });
    }
    saveIceCandidate(uuid, candidate) {
        const pairRecord = this.pairRecordMap.get(uuid);
        if (pairRecord) {
            pairRecord.get().then(doc => {
                const data = doc.data();
                if (data) {
                    const pairingRecord = JSON.parse(data.info);
                    if (uuid === pairingRecord.uuid1) {
                        pairingRecord.iceCandidates2.push(candidate);
                    }
                    else {
                        pairingRecord.iceCandidates1.push(candidate);
                    }
                    pairRecord.set({ info: JSON.stringify(pairingRecord) });
                }
            });
        }
    }
    savePairing(mesh, myIndex, index, peer) {
        if (this.roomRecord) {
            const pairRecordName = `pair_${this.roomRecord.roomId}_${Math.min(index, myIndex)}_${Math.max(index, myIndex)}`;
            const pairRecord = this.roomsCollectionRef.doc(pairRecordName);
            this.pairRecordMap.set(peer.uuid, pairRecord);
            pairRecord.onSnapshot((doc) => {
                const data = doc.data();
                if (data) {
                    this.pairingRecordChanged(mesh, JSON.parse(data.info), peer, myIndex, index);
                }
                else if (myIndex < index) {
                    const pairingRecord = new PairingRecord(0, '', '', [], [], '', '');
                    pairRecord.set({ info: JSON.stringify(pairingRecord) });
                }
            });
        }
    }
    savePairingRecord(info, uuid) {
        const pairRecord = this.pairRecordMap.get(uuid);
        if (pairRecord)
            pairRecord.set({ info: JSON.stringify(info) });
    }
}

export { FirebaseSignaling };
