declare module "event-emitter" {
    export class EventEmitter<T> {
        private _listeners;
        constructor();
        addEventListener(listener: (uuid: string, event: T) => void): void;
        removeEventListener(listener: (uuid: string, event: T) => void): void;
        notify(uuid: string, event: T): void;
    }
}
declare module "mesh-config" {
    export class MeshConfig {
        private _rtcConfig;
        private _rtcDataChannelInit;
        private _messagesAwaitingReplyMaxSize;
        private _messagesAwaitingReplyMaxAge;
        private _messagesAwaitingReplyCleanerInterval;
        private _checkLatencyInterval;
        constructor(_rtcConfig: RTCConfiguration, _rtcDataChannelInit: RTCDataChannelInit, _messagesAwaitingReplyMaxSize: number, _messagesAwaitingReplyMaxAge: number, _messagesAwaitingReplyCleanerInterval: number, _checkLatencyInterval: number);
        get rtcConfig(): RTCConfiguration;
        get rtcDataChannelInit(): RTCDataChannelInit;
        get messagesAwaitingReplyMaxSize(): number;
        get messagesAwaitingReplyMaxAge(): number;
        get messagesAwaitingReplyCleanerInterval(): number;
        get checkLatencyInterval(): number;
    }
}
declare module "message" {
    /*************************************
     * Message header:
     *   - timestamp: 8 bytes
     *   - sequence: 2 bytes (0..65535)
     *   - type: 1 byte (1 - send ; 2 - sendAndListen ; 3 - reply ; 4 - replyAndListen)
     *   - if (type == 3 || type == 4)
     *     - source timestamp: 8 bytes
     *     - source sequence: 2 bytes
     *
     * SYSTEM MESSAGES
     *   - body = [1] => ping
     *
     *
     * https://javascript.info/arraybuffer-binary-arrays
     *************************************/
    export const MESSAGE_SEND = 1;
    export const MESSAGE_SEND_AND_LISTEN = 2;
    export const MESSAGE_REPLY = 3;
    export const MESSAGE_REPLY_AND_LISTEN = 4;
    export const SYSTEM_MESSAGE_PING = 1;
    export const SYSTEM_MESSAGE_IM_READY = 2;
    export class Message {
        private _timestamp;
        private _sequence;
        private _type;
        private _sourceTimestamp?;
        private _sourceSequence?;
        private _body;
        private _fullMessage;
        private _clockDiff;
        static parse(fullMessage: ArrayBuffer): Message;
        constructor(body: ArrayBuffer, timestamp: number, sequence: number, type: number, sourceTimestamp?: number, _sourceSequence?: number);
        get timestamp(): number;
        get sequence(): number;
        get type(): number;
        get sourceTimestamp(): number | undefined;
        get sourceSequence(): number | undefined;
        get body(): ArrayBuffer;
        get fullMessage(): ArrayBuffer;
        get key(): string;
        get sourceKey(): string;
        get awaitReply(): boolean;
        get clockDiff(): number | undefined;
        set clockDiff(value: number | undefined);
        get timestampToLocalTime(): number | undefined;
    }
}
declare module "utils" {
    let MPLAYNET_DEBUG: boolean;
    export { MPLAYNET_DEBUG };
    export function setDebug(value: boolean): void;
    export function uuidv4(): string;
    export function getLocalTimestamp(): number;
}
declare module "mesh-connection" {
    import { EventEmitter } from "event-emitter";
    import { MeshConfig } from "mesh-config";
    import { Message } from "message";
    export class MeshConnection {
        private config;
        private _uuid;
        private static SYSTEM_CHANNEL_LABEL;
        private static APP_CHANNEL_LABEL;
        private _connection;
        private _channel;
        private _systemChannel;
        private _connectionReady;
        private _connectionReadyEmitter;
        private _iceCandidateEmitter;
        private _messageEmitter;
        private _systemMessageEmitter;
        private _messageSeq;
        private _messagesAwaitingReply;
        private _messagesAwaitingReplyCount;
        private _messagesAwaitingReplyCleanerTimer;
        private _latency;
        private _minLatency;
        private _clockDiff;
        private _checkLatencyIntervalTimer;
        constructor(config: MeshConfig, _uuid: string);
        addIceCandidate(candidate: RTCIceCandidateInit): void;
        get uuid(): string;
        get connectionReadyEmitter(): EventEmitter<boolean>;
        get iceCandidateEmitter(): EventEmitter<RTCIceCandidate>;
        get messageEmitter(): EventEmitter<Message>;
        get systemMessageEmitter(): EventEmitter<Message>;
        get isOpen(): boolean;
        get latency(): number | undefined;
        get clockDiff(): number | undefined;
        close(): void;
        createOffer(): Promise<RTCSessionDescriptionInit>;
        createAnswer(offer: RTCSessionDescriptionInit): Promise<RTCSessionDescriptionInit>;
        verifyAnswer(answer: RTCSessionDescriptionInit): Promise<void>;
        private updateLatencyAndClockDiff;
        private connectionReadyTasks;
        private sendPing;
        private emitConnectionIsNotReady;
        private cleanMessagesAwaitingReply;
        private storeMessagesAwaitingReply;
        private processReceivedMessage;
        private internalSend;
        private sendByChannel;
        private sendAndListenBychannel;
        private replyByChannel;
        private replyAndListenByChannel;
        send(message: ArrayBuffer): boolean;
        sendSys(message: ArrayBuffer): boolean;
        sendAndListen(message: ArrayBuffer): Promise<Message>;
        sendAndListenSys(message: ArrayBuffer): Promise<Message>;
        reply(originalMessage: Message, message: ArrayBuffer): boolean;
        replySys(originalMessage: Message, message: ArrayBuffer): boolean;
        replyAndListen(originalMessage: Message, message: ArrayBuffer): Promise<Message>;
        replyAndListenSys(originalMessage: Message, message: ArrayBuffer): Promise<Message>;
    }
}
declare module "mesh" {
    import { EventEmitter } from "event-emitter";
    import { MeshConfig } from "mesh-config";
    import { Message } from "message";
    export class Mesh {
        private config;
        private _uuid;
        private _connections;
        private _meshReadyEmitter;
        private _connectionReadyEmitter;
        private _iceCandidateEmitter;
        private _messageEmitter;
        private _peerReadyCount;
        constructor(config: MeshConfig, _uuid: string);
        get uuid(): string;
        get meshReadyEmitter(): EventEmitter<boolean>;
        get connectionReadyEmitter(): EventEmitter<boolean>;
        get iceCandidateEmitter(): EventEmitter<RTCIceCandidate>;
        get messageEmitter(): EventEmitter<Message>;
        createConnection(uuid: string): void;
        close(): void;
        connectionCount(): number;
        connectionsOpened(): number;
        private findConnection;
        addIceCandidate(uuid: string, candidate: RTCIceCandidateInit): void;
        createOffer(uuid: string): Promise<RTCSessionDescriptionInit>;
        createAnswer(uuid: string, offer: RTCSessionDescriptionInit): Promise<RTCSessionDescriptionInit>;
        verifyAnswer(uuid: string, answer: RTCSessionDescriptionInit): Promise<void>;
        send(uuid: string, message: ArrayBuffer): boolean;
        sendAndListen(uuid: string, message: ArrayBuffer): Promise<Message>;
        reply(uuid: string, originalMessage: Message, message: ArrayBuffer): boolean;
        replyAndListen(uuid: string, originalMessage: Message, message: ArrayBuffer): Promise<Message>;
        broadcast(message: ArrayBuffer): number;
        broadcastAndListen(message: ArrayBuffer): Promise<Message>[];
        private broadcastSys;
        private systemMessageArrived;
    }
}
declare module "signaling/records" {
    export class PeerRecord {
        username: string;
        uuid: string;
        ready: boolean;
        constructor(username: string, uuid: string, ready: boolean);
    }
    export class RoomRecord {
        roomId: string;
        peers: PeerRecord[];
        constructor(roomId: string, peers: PeerRecord[]);
    }
    export class PairingRecord {
        status: number;
        uuid1: string;
        uuid2: string;
        iceCandidates1: string[];
        iceCandidates2: string[];
        offer: string;
        answer: string;
        constructor(status: number, uuid1: string, uuid2: string, iceCandidates1: string[], iceCandidates2: string[], offer: string, answer: string);
    }
    export class PairingItem {
        iceCandidatesDone: string[];
        answered: boolean;
        answerVerified: boolean;
        constructor(iceCandidatesDone: string[], answered: boolean, answerVerified: boolean);
    }
}
declare module "signaling/base-signaling" {
    import { EventEmitter } from "event-emitter";
    import { Mesh } from "mesh";
    import { PairingItem, PairingRecord, PeerRecord, RoomRecord } from "signaling/records";
    export abstract class BaseSignaling {
        protected _uuid: string;
        protected roomRecord: RoomRecord | null;
        protected _roomRecordEmitter: EventEmitter<RoomRecord>;
        protected pairingMap: Map<string, PairingItem>;
        get roomRecordEmitter(): EventEmitter<RoomRecord>;
        get uuid(): string;
        constructor();
        hostRoom(roomId: string, username: string, uuid: string): Promise<boolean>;
        joinRoom(roomId: string, username: string, uuid: string): Promise<boolean>;
        protected abstract internalHostRoom(roomId: string, username: string): Promise<boolean>;
        protected abstract internalJoinRoom(roomId: string, username: string): Promise<boolean>;
        protected abstract saveRoomInfo(): void;
        protected abstract saveIceCandidate(uuid: string, candidate: string): void;
        protected abstract savePairing(mesh: Mesh, myIndex: number, index: number, peer: PeerRecord): void;
        protected abstract savePairingRecord(info: PairingRecord, uuid: string): void;
        protected abstract cleanup(): void;
        upatePlayerStatus(ready: boolean): void;
        protected roomRecordChanged(info: RoomRecord, roomId: string, username: string, uuid: string): void;
        startPairings(mesh: Mesh): Promise<boolean>;
        protected pairingRecordChanged(mesh: Mesh, info: PairingRecord, peer: PeerRecord, myIndex: number, index: number): void;
    }
}
declare module "signaling/local-signaling" {
    import { Mesh } from "mesh";
    import { BaseSignaling } from "signaling/base-signaling";
    import { PeerRecord, PairingRecord } from "signaling/records";
    export class LocalSignaling extends BaseSignaling {
        private roomId;
        private username;
        private pairRecordMap;
        private timers;
        constructor();
        protected cleanup(): void;
        private listenRoomRecordChanges;
        protected internalHostRoom(roomId: string, username: string): Promise<boolean>;
        protected internalJoinRoom(roomId: string, username: string): Promise<boolean>;
        protected saveRoomInfo(): void;
        protected saveIceCandidate(uuid: string, candidate: string): void;
        protected savePairing(mesh: Mesh, myIndex: number, index: number, peer: PeerRecord): void;
        protected savePairingRecord(info: PairingRecord, uuid: string): void;
    }
}
declare module "mplaynet" {
    import { Mesh } from "mesh";
    import { MeshConfig } from "mesh-config";
    import { LocalSignaling } from "signaling/local-signaling";
    import { PeerRecord } from "signaling/records";
    import { Message } from "message";
    import { getLocalTimestamp, setDebug } from "utils";
    export { Mesh, MeshConfig, PeerRecord, LocalSignaling, Message, getLocalTimestamp, setDebug };
}
