class EventEmitter {
    constructor() {
        this._listeners = [];
    }
    addEventListener(listener) {
        if (!this._listeners.includes(listener)) {
            this._listeners.push(listener);
        }
    }
    removeEventListener(listener) {
        const idx = this._listeners.indexOf(listener);
        if (idx !== -1) {
            this._listeners.splice(idx, 1);
        }
    }
    notify(uuid, event) {
        this._listeners.forEach((_listeners) => {
            _listeners(uuid, event);
        });
    }
}

/*************************************
 * Message header:
 *   - timestamp: 8 bytes
 *   - sequence: 2 bytes (0..65535)
 *   - type: 1 byte (1 - send ; 2 - sendAndListen ; 3 - reply ; 4 - replyAndListen)
 *   - if (type == 3 || type == 4)
 *     - source timestamp: 8 bytes
 *     - source sequence: 2 bytes
 *
 * SYSTEM MESSAGES
 *   - body = [1] => ping
 *
 *
 * https://javascript.info/arraybuffer-binary-arrays
 *************************************/
const MESSAGE_SEND = 1;
const MESSAGE_SEND_AND_LISTEN = 2;
const MESSAGE_REPLY = 3;
const MESSAGE_REPLY_AND_LISTEN = 4;
const SYSTEM_MESSAGE_PING = 1;
const SYSTEM_MESSAGE_IM_READY = 2;
class Message {
    constructor(body, timestamp, sequence, type, sourceTimestamp, _sourceSequence) {
        // body rececived: add headers to build fullMessage
        this._timestamp = timestamp;
        this._sequence = sequence;
        this._type = type;
        this._body = body;
        const isReply = (type === MESSAGE_REPLY || type === MESSAGE_REPLY_AND_LISTEN);
        // build full message
        this._fullMessage = new ArrayBuffer((!isReply ? 11 : 21) + this._body.byteLength);
        // set header
        // - timestamp
        new DataView(this._fullMessage).setBigInt64(0, BigInt(this._timestamp));
        // - sequence
        new DataView(this._fullMessage, 8).setInt16(0, this._sequence);
        // - type
        new DataView(this._fullMessage, 10).setInt8(0, this._type);
        if (isReply) {
            this._sourceTimestamp = sourceTimestamp;
            // - sourceTimestamp
            if (this._sourceTimestamp)
                new DataView(this._fullMessage, 11).setBigInt64(0, BigInt(this._sourceTimestamp));
            // - sequence
            this._sourceSequence = _sourceSequence;
            if (this._sourceSequence)
                new DataView(this._fullMessage, 19).setInt16(0, this._sourceSequence);
        }
        // set body
        new Uint8Array(this._fullMessage).set(new Uint8Array(this._body), !isReply ? 11 : 21);
    }
    static parse(fullMessage) {
        // extract headers and body from fullMessage
        let timestamp, sequence, type, sourceTimestamp, sourceSequence, body;
        // timestamp
        timestamp = Number(new DataView(fullMessage).getBigInt64(0));
        // sequence
        sequence = new DataView(fullMessage, 8).getInt16(0);
        // type
        type = new DataView(fullMessage, 10).getInt8(0);
        const isReply = (type === MESSAGE_REPLY || type === MESSAGE_REPLY_AND_LISTEN);
        // body
        body = fullMessage.slice(!isReply ? 11 : 21);
        if (isReply) {
            // - sourceTimestamp
            sourceTimestamp = Number(new DataView(fullMessage, 11).getBigInt64(0));
            // - sourceSequence
            sourceSequence = new DataView(fullMessage, 19).getInt16(0);
            return new Message(body, timestamp, sequence, type, sourceTimestamp, sourceSequence);
        }
        else {
            return new Message(body, timestamp, sequence, type);
        }
    }
    get timestamp() { return this._timestamp; }
    get sequence() { return this._sequence; }
    get type() { return this._type; }
    get sourceTimestamp() { return this._sourceTimestamp; }
    get sourceSequence() { return this._sourceSequence; }
    get body() { return this._body; }
    get fullMessage() { return this._fullMessage; }
    get key() { return `${this._timestamp}:${this._sequence}`; }
    get sourceKey() { return `${this._sourceTimestamp}:${this._sourceSequence}`; }
    get awaitReply() { return (this._type === MESSAGE_SEND_AND_LISTEN || this._type === MESSAGE_REPLY_AND_LISTEN); }
    get clockDiff() { return this._clockDiff; }
    set clockDiff(value) { this._clockDiff = value; }
    get timestampToLocalTime() { return this._clockDiff !== undefined ? this._timestamp + this._clockDiff : undefined; }
}

let MPLAYNET_DEBUG = false;
function setDebug(value) {
    MPLAYNET_DEBUG = value;
}
function uuidv4() {
    return 'xxxxxxxx-xxxx-4xxx-yxxx-xxxxxxxxxxxx'.replace(/[xy]/g, function (c) {
        let r = (Math.random() * 16) | 0, v = c == 'x' ? r : (r & 0x3) | 0x8;
        return v.toString(16);
    });
}
const FAKE_TIME = // for local debugging
 0;
/*
  10000 * (1 + Math.floor(Math.random() * 9)) +
  1000 * (1 + Math.floor(Math.random() * 9)) +
    100 * (1 + Math.floor(Math.random() * 9));
//*/
function getLocalTimestamp() {
    return new Date().valueOf() + FAKE_TIME;
}
// for debugging purposes
if (MPLAYNET_DEBUG) {
    const fakeTimestamp = getLocalTimestamp();
    console.log(`FAKE TIME: ${FAKE_TIME}, fake timestamp=${fakeTimestamp}, time=${fakeTimestamp - FAKE_TIME}`);
}

class MeshConnection {
    constructor(config, _uuid) {
        this.config = config;
        this._uuid = _uuid;
        this._connectionReady = false;
        this._messageSeq = 1;
        // cache with messages awaiting reply
        this._messagesAwaitingReply = new Map();
        this._messagesAwaitingReplyCount = 0;
        this._messagesAwaitingReplyCleanerTimer = null;
        this._checkLatencyIntervalTimer = null;
        getLocalTimestamp();
        this._connectionReadyEmitter = new EventEmitter();
        this._iceCandidateEmitter = new EventEmitter();
        this._messageEmitter = new EventEmitter();
        this._systemMessageEmitter = new EventEmitter();
        this._connection = new RTCPeerConnection(this.config.rtcConfig);
        this._connection.onicecandidate = (event) => {
            if (event.candidate)
                this.iceCandidateEmitter.notify(this._uuid, event.candidate);
        };
        //this._connection.onconnectionstatechange = () => {
        //  console.log(`connectionState?${this._connection.connectionState}`);
        //};
        this._systemChannel = this._connection.createDataChannel(MeshConnection.SYSTEM_CHANNEL_LABEL, { ordered: true });
        this._channel = this._connection.createDataChannel(MeshConnection.APP_CHANNEL_LABEL, this.config.rtcDataChannelInit);
        this._channel.onopen = () => {
            if (!this._connectionReady) {
                this._connectionReady = true;
            }
            else {
                this.connectionReadyTasks();
            }
        };
        this._channel.onclose = () => {
            if (this._connectionReady) {
                this._connectionReady = false;
            }
            else {
                this.emitConnectionIsNotReady();
            }
        };
        this._systemChannel.onopen = () => {
            if (!this._connectionReady) {
                this._connectionReady = true;
            }
            else {
                this.connectionReadyTasks();
            }
        };
        this._systemChannel.onclose = () => {
            if (this._connectionReady) {
                this._connectionReady = false;
            }
            else {
                this.emitConnectionIsNotReady();
            }
        };
        this._connection.ondatachannel = (ev) => {
            ev.channel.binaryType = "arraybuffer";
            switch (ev.channel.label) {
                case MeshConnection.SYSTEM_CHANNEL_LABEL:
                    ev.channel.onmessage = (evm) => {
                        this.processReceivedMessage(this._systemMessageEmitter, Message.parse(evm.data), true);
                    };
                    break;
                case MeshConnection.APP_CHANNEL_LABEL:
                    ev.channel.onmessage = (evm) => {
                        this.processReceivedMessage(this._messageEmitter, Message.parse(evm.data), false);
                    };
                    break;
            }
        };
    }
    addIceCandidate(candidate) {
        this._connection.addIceCandidate(candidate);
    }
    get uuid() {
        return this._uuid;
    }
    get connectionReadyEmitter() {
        return this._connectionReadyEmitter;
    }
    get iceCandidateEmitter() {
        return this._iceCandidateEmitter;
    }
    get messageEmitter() {
        return this._messageEmitter;
    }
    get systemMessageEmitter() {
        return this._systemMessageEmitter;
    }
    get isOpen() {
        return this._channel && this._channel.readyState === 'open' &&
            this._systemChannel && this._systemChannel.readyState === 'open';
    }
    get latency() {
        return this._latency;
    }
    get clockDiff() {
        return this._clockDiff;
    }
    close() {
        if (this._checkLatencyIntervalTimer)
            window.clearInterval(this._checkLatencyIntervalTimer);
        if (this._messagesAwaitingReplyCleanerTimer)
            window.clearInterval(this._messagesAwaitingReplyCleanerTimer);
        this._messagesAwaitingReply.clear();
        this._messagesAwaitingReplyCount = 0;
        if (this._systemChannel)
            try {
                this._systemChannel.close();
            }
            catch (err) { }
        if (this._channel)
            try {
                this._channel.close();
            }
            catch (err) { }
        if (this._connection)
            try {
                this._connection.close();
            }
            catch (err) { }
    }
    createOffer() {
        return this._connection.createOffer().then((offer) => {
            //if (offer && offer.sdp) offer.sdp = offer.sdp.replace(/a=ice-options:trickle\s\n/g, '');
            return this._connection.setLocalDescription(offer).then(() => {
                return offer;
            });
        });
    }
    createAnswer(offer) {
        return this._connection
            .setRemoteDescription(offer)
            .then(() => this._connection.createAnswer())
            .then((answer) => {
            //if (answer && answer.sdp) answer.sdp = answer.sdp.replace(/a=ice-options:trickle\s\n/g, '');
            return this._connection.setLocalDescription(answer).then(() => {
                return answer;
            });
        });
    }
    verifyAnswer(answer) {
        return this._connection.setRemoteDescription(answer);
    }
    // Cristian's algorithm (https://en.wikipedia.org/wiki/Cristian%27s_algorithm)
    updateLatencyAndClockDiff(currentLocalTimestamp, previousLocalTimestamp, remoteTimestamp) {
        this._latency = currentLocalTimestamp - previousLocalTimestamp;
        const updateClock = (!this._minLatency || this._latency <= this._minLatency);
        if (updateClock) {
            this._minLatency = this._latency;
            const lag = Math.round(this._latency / 2);
            this._clockDiff = currentLocalTimestamp - lag - remoteTimestamp;
        }
        if (MPLAYNET_DEBUG)
            console.log(`latency: ${this._latency}, clock diff=${this._clockDiff}`);
    }
    connectionReadyTasks() {
        // ping at regular intervals to update latency and clock differences
        this.sendPing();
        // TO DO: check if this can be done in a webworker
        this._checkLatencyIntervalTimer = window.setInterval(() => { this.sendPing(); }, this.config.checkLatencyInterval);
        // launch timer to clean messages awaiting reply // TO DO: check if this can be done in a webworker
        if (this.config.messagesAwaitingReplyMaxAge > 0) {
            this._messagesAwaitingReplyCleanerTimer = window.setInterval(() => { this.cleanMessagesAwaitingReply(); }, this.config.messagesAwaitingReplyCleanerInterval);
        }
        // notify connection is ready
        this._connectionReadyEmitter.notify(this._uuid, true);
    }
    sendPing() {
        const packet = new Uint8Array(1);
        packet[0] = SYSTEM_MESSAGE_PING;
        this.sendAndListenSys(packet.buffer).then(reply => {
            if (reply.sourceTimestamp)
                this.updateLatencyAndClockDiff(getLocalTimestamp(), reply.sourceTimestamp, reply.timestamp);
        });
    }
    emitConnectionIsNotReady() {
        this._connectionReadyEmitter.notify(this._uuid, false);
    }
    cleanMessagesAwaitingReply() {
        const time = getLocalTimestamp() - this.config.messagesAwaitingReplyMaxAge;
        for (let key of this._messagesAwaitingReply.keys()) {
            if (+key.split(':')[0] < time) {
                this._messagesAwaitingReply.delete(key);
                this._messagesAwaitingReplyCount--;
            }
            else {
                break;
            }
        }
    }
    storeMessagesAwaitingReply(key, value) {
        this._messagesAwaitingReply.set(key, value);
        if (this.config.messagesAwaitingReplyMaxSize > 0 && this._messagesAwaitingReplyCount >= this.config.messagesAwaitingReplyMaxSize) {
            // remove first element
            this._messagesAwaitingReply.delete(this._messagesAwaitingReply.entries().next().value[0]);
        }
        else {
            this._messagesAwaitingReplyCount++;
        }
    }
    processReceivedMessage(emmiter, message, isSystemMessage) {
        const now = getLocalTimestamp();
        // if it is a system message, process it accordingly
        if (isSystemMessage) {
            // check if the message requires reply
            if (message.type === MESSAGE_SEND_AND_LISTEN) {
                const data = new Int8Array(message.body);
                // if PING, send reply
                if (data[0] === SYSTEM_MESSAGE_PING) {
                    this.replySys(message, data.buffer);
                    return; // do not want to bubble up the ping event
                }
            }
        }
        message.clockDiff = this._clockDiff;
        // if it is a reply, look for the original message in the message cache
        if (message.type === MESSAGE_REPLY || message.type === MESSAGE_REPLY_AND_LISTEN) {
            if (MPLAYNET_DEBUG)
                (`message received at localTimestamp ${now}: timestamp=${message.timestamp} (toLocalTime=${message.timestampToLocalTime}), sequence=${message.sequence}, sourceTimestamp=${message.sourceTimestamp}, sourceSequence=${message.sourceSequence}`);
            const key = message.sourceKey;
            const sourceMessage = this._messagesAwaitingReply.get(key);
            if (sourceMessage) {
                this._messagesAwaitingReply.delete(key);
                this._messagesAwaitingReplyCount--;
                sourceMessage(message);
                return;
            }
        }
        else {
            if (MPLAYNET_DEBUG)
                console.log(`message received at localTimestamp ${now}: timestamp=${message.timestamp} (toLocalTime=${message.timestampToLocalTime}), sequence=${message.sequence}`);
        }
        emmiter.notify(this._uuid, message);
    }
    internalSend(channel, message) {
        if (channel.readyState !== 'open') {
            return false;
        }
        channel.send(message);
        return true;
    }
    sendByChannel(channel, message) {
        const timestamp = getLocalTimestamp();
        const theMessage = new Message(message, timestamp, this._messageSeq++, MESSAGE_SEND);
        if (MPLAYNET_DEBUG)
            console.log(`sending timestamp=${theMessage.timestamp}, sequence=${theMessage.sequence}`);
        return this.internalSend(channel, theMessage.fullMessage);
    }
    sendAndListenBychannel(channel, message) {
        return new Promise(resolve => {
            const timestamp = getLocalTimestamp();
            const theMessage = new Message(message, timestamp, this._messageSeq++, MESSAGE_SEND_AND_LISTEN);
            if (MPLAYNET_DEBUG)
                console.log(`sending timestamp=${theMessage.timestamp}, sequence=${theMessage.sequence}`);
            if (this.internalSend(channel, theMessage.fullMessage)) {
                // CAUTION: unlikely race condition if reply arrives before message is cached
                this.storeMessagesAwaitingReply(theMessage.key, resolve);
            }
        });
    }
    replyByChannel(channel, originalMessage, message) {
        const timestamp = getLocalTimestamp();
        const theMessage = new Message(message, timestamp, this._messageSeq++, MESSAGE_REPLY, originalMessage.timestamp, originalMessage.sequence);
        if (MPLAYNET_DEBUG)
            console.log(`sending timestamp=${theMessage.timestamp}, sequence=${theMessage.sequence}, sourceTimestamp=${originalMessage.timestamp}, sourceSequence=${originalMessage.sequence}`);
        return this.internalSend(channel, theMessage.fullMessage);
    }
    replyAndListenByChannel(channel, originalMessage, message) {
        return new Promise(resolve => {
            const timestamp = getLocalTimestamp();
            const theMessage = new Message(message, timestamp, this._messageSeq++, MESSAGE_REPLY_AND_LISTEN, originalMessage.timestamp, originalMessage.sequence);
            if (MPLAYNET_DEBUG)
                console.log(`sending timestamp=${theMessage.timestamp}, sequence=${theMessage.sequence}, sourceTimestamp=${originalMessage.timestamp}, sourceSequence=${originalMessage.sequence}`);
            if (this.internalSend(channel, theMessage.fullMessage)) {
                // CAUTION: unlikely race condition if reply arrives before message is cached
                this.storeMessagesAwaitingReply(theMessage.key, resolve);
            }
        });
    }
    send(message) {
        return this.sendByChannel(this._channel, message);
    }
    sendSys(message) {
        return this.sendByChannel(this._systemChannel, message);
    }
    sendAndListen(message) {
        return this.sendAndListenBychannel(this._channel, message);
    }
    sendAndListenSys(message) {
        return this.sendAndListenBychannel(this._systemChannel, message);
    }
    reply(originalMessage, message) {
        return this.replyByChannel(this._channel, originalMessage, message);
    }
    replySys(originalMessage, message) {
        return this.replyByChannel(this._systemChannel, originalMessage, message);
    }
    replyAndListen(originalMessage, message) {
        return this.replyAndListenByChannel(this._channel, originalMessage, message);
    }
    replyAndListenSys(originalMessage, message) {
        return this.replyAndListenByChannel(this._systemChannel, originalMessage, message);
    }
}
MeshConnection.SYSTEM_CHANNEL_LABEL = 'system-channel';
MeshConnection.APP_CHANNEL_LABEL = 'app-channel';

class Mesh {
    constructor(config, _uuid) {
        this.config = config;
        this._uuid = _uuid;
        this._connections = [];
        this._meshReadyEmitter = new EventEmitter();
        this._connectionReadyEmitter = new EventEmitter();
        this._iceCandidateEmitter = new EventEmitter();
        this._messageEmitter = new EventEmitter();
        this._peerReadyCount = 0;
    }
    get uuid() {
        return this._uuid;
    }
    get meshReadyEmitter() {
        return this._meshReadyEmitter;
    }
    get connectionReadyEmitter() {
        return this._connectionReadyEmitter;
    }
    get iceCandidateEmitter() {
        return this._iceCandidateEmitter;
    }
    get messageEmitter() {
        return this._messageEmitter;
    }
    createConnection(uuid) {
        if (MPLAYNET_DEBUG)
            console.log(`createConnection ${uuid}`);
        const connection = new MeshConnection(this.config, uuid);
        connection.connectionReadyEmitter.addEventListener((uuid, ready) => {
            if (!ready) {
                const connIdx = this._connections.findIndex(conn => conn.uuid === uuid);
                if (connIdx >= 0) {
                    this._connections[connIdx].close();
                    this._connections.splice(connIdx, 1);
                }
            }
            else {
                // If I have already established a connection with all the peers, notify them that I am ready
                if (this.connectionCount() === this.connectionsOpened()) {
                    if (MPLAYNET_DEBUG)
                        console.log('I am ready!!!!!!!!!');
                    const packet = new Uint8Array(1);
                    packet[0] = SYSTEM_MESSAGE_IM_READY;
                    this.broadcastSys(packet.buffer);
                }
            }
            this._connectionReadyEmitter.notify(uuid, ready);
        });
        connection.iceCandidateEmitter.addEventListener((uuid, event) => {
            this._iceCandidateEmitter.notify(uuid, event);
        });
        connection.messageEmitter.addEventListener((uuid, event) => {
            this._messageEmitter.notify(uuid, event);
        });
        connection.systemMessageEmitter.addEventListener((uuid, event) => {
            this.systemMessageArrived(uuid, event);
        });
        this._connections.push(connection);
    }
    close() {
        this._connections.forEach((conn) => {
            conn.close();
        });
    }
    connectionCount() {
        return this._connections.length;
    }
    connectionsOpened() {
        return this._connections.reduce((total, conn) => (conn.isOpen ? ++total : total), 0);
    }
    findConnection(uuid) {
        return this._connections.find((conn) => conn.uuid === uuid);
    }
    addIceCandidate(uuid, candidate) {
        const conn = this.findConnection(uuid);
        if (conn)
            return conn.addIceCandidate(candidate);
        else
            throw new Error(`meshConnection ${uuid} not found`);
    }
    createOffer(uuid) {
        const conn = this.findConnection(uuid);
        if (conn)
            return conn.createOffer();
        else
            throw new Error(`meshConnection ${uuid} not found`);
    }
    createAnswer(uuid, offer) {
        const conn = this.findConnection(uuid);
        if (conn)
            return conn.createAnswer(offer);
        else
            throw new Error(`meshConnection ${uuid} not found`);
    }
    verifyAnswer(uuid, answer) {
        const conn = this.findConnection(uuid);
        if (conn)
            return conn.verifyAnswer(answer);
        else
            throw new Error(`meshConnection ${uuid} not found`);
    }
    send(uuid, message) {
        const conn = this.findConnection(uuid);
        if (conn)
            return conn.send(message);
        else
            throw new Error(`meshConnection ${uuid} not found`);
    }
    sendAndListen(uuid, message) {
        const conn = this.findConnection(uuid);
        if (conn)
            return conn.sendAndListen(message);
        else
            throw new Error(`meshConnection ${uuid} not found`);
    }
    reply(uuid, originalMessage, message) {
        const conn = this.findConnection(uuid);
        if (conn)
            return conn.reply(originalMessage, message);
        else
            throw new Error(`meshConnection ${uuid} not found`);
    }
    replyAndListen(uuid, originalMessage, message) {
        const conn = this.findConnection(uuid);
        if (conn)
            return conn.replyAndListen(originalMessage, message);
        else
            throw new Error(`meshConnection ${uuid} not found`);
    }
    broadcast(message) {
        return this._connections.reduce((total, conn) => (conn.send(message) ? ++total : total), 0);
    }
    broadcastAndListen(message) {
        return this._connections.map(conn => conn.sendAndListen(message));
    }
    broadcastSys(message) {
        return this._connections.reduce((total, conn) => (conn.sendSys(message) ? ++total : total), 0);
    }
    systemMessageArrived(uuid, message) {
        const data = new Int8Array(message.body);
        if (data[0] === SYSTEM_MESSAGE_IM_READY) {
            if (MPLAYNET_DEBUG)
                console.log(`${uuid} is ready!!!!!!!`);
            if (++this._peerReadyCount >= this._connections.length) {
                if (MPLAYNET_DEBUG)
                    console.log('All peers are ready!!!!!!!!!!');
                this._meshReadyEmitter.notify(this._uuid, true);
            }
        }
    }
}

class MeshConfig {
    constructor(
    /*
    The RTCConfiguration defines a set of parameters to configure how the peer-to-peer communication
    established via RTCPeerConnection is established or re-established.
    
    interface RTCConfiguration {
        iceServers?: RTCIceServer[];
        iceTransportPolicy?: RTCIceTransportPolicy; // default = 'all'
        bundlePolicy?: RTCBundlePolicy; // default = 'balanced'
        rtcpMuxPolicy?: RTCRtcpMuxPolicy; // default = 'require'
        peerIdentity?: string; // default = null
        certificates?: RTCCertificate[];
        iceCandidatePoolSize?: number; // default = 0
    }
    */
    _rtcConfig, 
    /*
    An RTCDataChannel can be configured to operate in different reliability modes.
    A reliable channel ensures that the data is delivered at the other peer through retransmissions.
    An unreliable channel is configured to either limit the number of retransmissions
    ( maxRetransmits ) or set a time during which transmissions (including retransmissions)
    are allowed ( maxPacketLifeTime ).
    These properties can not be used simultaneously and an attempt to do so will result in an error.
    Not setting any of these properties results in a reliable channel.
    
    interface RTCDataChannelInit {
        ordered?: boolean; // default = true
        maxPacketLifeTime?: number;
        maxRetransmits?: number;
        protocol?: string; // default = ''
        negotiated?: boolean; // default = false
        id?: number;
    }
    */
    _rtcDataChannelInit, 
    /* Max number of messages awaiting reply, or 0 for infinite */
    _messagesAwaitingReplyMaxSize, 
    /* Max age for messages awaiting reply, or 0 so they do not expire */
    _messagesAwaitingReplyMaxAge, 
    /* Interval (milliseconds) for cleaning the buffer of messages awaiting reply */
    _messagesAwaitingReplyCleanerInterval, 
    /* Interval (milliseconds) to calculate latency and synchronize clocks */
    _checkLatencyInterval) {
        this._rtcConfig = _rtcConfig;
        this._rtcDataChannelInit = _rtcDataChannelInit;
        this._messagesAwaitingReplyMaxSize = _messagesAwaitingReplyMaxSize;
        this._messagesAwaitingReplyMaxAge = _messagesAwaitingReplyMaxAge;
        this._messagesAwaitingReplyCleanerInterval = _messagesAwaitingReplyCleanerInterval;
        this._checkLatencyInterval = _checkLatencyInterval;
    }
    get rtcConfig() {
        return this._rtcConfig;
    }
    get rtcDataChannelInit() {
        return this._rtcDataChannelInit;
    }
    get messagesAwaitingReplyMaxSize() {
        return this._messagesAwaitingReplyMaxSize;
    }
    get messagesAwaitingReplyMaxAge() {
        return this._messagesAwaitingReplyMaxAge;
    }
    get messagesAwaitingReplyCleanerInterval() {
        return this._messagesAwaitingReplyCleanerInterval;
    }
    get checkLatencyInterval() {
        return this._checkLatencyInterval;
    }
}

class PeerRecord {
    constructor(username, uuid, ready) {
        this.username = username;
        this.uuid = uuid;
        this.ready = ready;
    }
}
class RoomRecord {
    constructor(roomId, peers) {
        this.roomId = roomId;
        this.peers = peers;
    }
}
class PairingRecord {
    constructor(status, uuid1, uuid2, iceCandidates1, iceCandidates2, offer, answer) {
        this.status = status;
        this.uuid1 = uuid1;
        this.uuid2 = uuid2;
        this.iceCandidates1 = iceCandidates1;
        this.iceCandidates2 = iceCandidates2;
        this.offer = offer;
        this.answer = answer;
    }
}
class PairingItem {
    constructor(iceCandidatesDone, answered, answerVerified) {
        this.iceCandidatesDone = iceCandidatesDone;
        this.answered = answered;
        this.answerVerified = answerVerified;
    }
}

class BaseSignaling {
    constructor() {
        this._uuid = '';
        this.roomRecord = null;
        this.pairingMap = new Map();
        this._roomRecordEmitter = new EventEmitter();
    }
    get roomRecordEmitter() {
        return this._roomRecordEmitter;
    }
    get uuid() {
        return this._uuid;
    }
    hostRoom(roomId, username, uuid) {
        if (!uuid)
            uuid = uuidv4();
        this._uuid = uuid;
        return this.internalHostRoom(roomId, username);
    }
    joinRoom(roomId, username, uuid) {
        if (!uuid)
            uuid = uuidv4();
        this._uuid = uuid;
        return this.internalJoinRoom(roomId, username);
    }
    upatePlayerStatus(ready) {
        if (this.roomRecord) {
            const peer = this.roomRecord.peers.find((peer) => peer.uuid === this._uuid);
            if (peer) {
                peer.ready = ready;
                this.saveRoomInfo();
            }
        }
    }
    roomRecordChanged(info, roomId, username, uuid) {
        if (!info || !info.peers) {
            this.roomRecord = new RoomRecord(roomId, []);
        }
        else {
            this.roomRecord = info;
        }
        if (!this.roomRecord.peers.some((peer) => peer.uuid === uuid)) {
            this.roomRecord.peers.push(new PeerRecord(username, uuid, false));
            this.saveRoomInfo();
        }
        else {
            this._roomRecordEmitter.notify(uuid, info);
        }
    }
    startPairings(mesh) {
        let pairing = true;
        return new Promise((resolve) => {
            mesh.meshReadyEmitter.addEventListener((_uuid, ready) => {
                this.cleanup();
                resolve(ready);
            });
            mesh.connectionReadyEmitter.addEventListener((_uuid, _ready) => {
                if (!pairing) {
                    return;
                }
                const connectionCount = mesh.connectionCount();
                const connectionsOpened = mesh.connectionsOpened();
                if (MPLAYNET_DEBUG)
                    console.log(`opened ${connectionsOpened} of ${connectionCount} connections`);
                if (connectionsOpened === connectionCount) {
                    pairing = false;
                }
            });
            mesh.iceCandidateEmitter.addEventListener((uuid, candidate) => {
                let sCandidate = JSON.stringify(candidate);
                if (!candidate.candidate)
                    return;
                if (MPLAYNET_DEBUG)
                    console.log(`candidate: ${sCandidate}`);
                this.saveIceCandidate(uuid, sCandidate);
            });
            if (this.roomRecord) {
                const myIndex = this.roomRecord.peers.findIndex((peer) => peer.uuid === this._uuid);
                for (let [index, peer] of this.roomRecord.peers.entries()) {
                    if (index === myIndex) {
                        continue;
                    }
                    this.savePairing(mesh, myIndex, index, peer);
                }
            }
        });
    }
    pairingRecordChanged(mesh, info, peer, myIndex, index) {
        if (!info || !info.status) {
            if (myIndex < index) {
                mesh.createConnection(peer.uuid);
                setTimeout(() => {
                    mesh.createOffer(peer.uuid).then((offer) => {
                        if (MPLAYNET_DEBUG)
                            console.log('create offer');
                        info = new PairingRecord(1, this._uuid, peer.uuid, [], [], JSON.stringify(offer), '');
                        this.pairingMap.set(peer.uuid, new PairingItem([], false, false));
                        this.savePairingRecord(info, peer.uuid);
                    });
                }, 0);
            }
        }
        else {
            let pairingItem = this.pairingMap.get(peer.uuid);
            if (!pairingItem) {
                pairingItem = new PairingItem([], false, false);
                this.pairingMap.set(peer.uuid, pairingItem);
            }
            if (myIndex > index && info.offer && !pairingItem.answered) {
                pairingItem.answered = true;
                mesh.createConnection(peer.uuid);
                mesh
                    .createAnswer(peer.uuid, JSON.parse(info.offer))
                    .then((answer) => {
                    if (MPLAYNET_DEBUG)
                        console.log('create answer');
                    info.answer = JSON.stringify(answer);
                    this.savePairingRecord(info, peer.uuid);
                });
            }
            if (myIndex < index &&
                info.answer &&
                !pairingItem.answerVerified) {
                if (MPLAYNET_DEBUG)
                    console.log('verify answer');
                pairingItem.answerVerified = true;
                mesh.verifyAnswer(peer.uuid, JSON.parse(info.answer));
            }
            let candidates;
            if (myIndex < index) {
                candidates = info.iceCandidates2;
            }
            else {
                candidates = info.iceCandidates1;
            }
            candidates.forEach((candidate) => {
                if (pairingItem &&
                    !pairingItem.iceCandidatesDone.includes(candidate)) {
                    pairingItem.iceCandidatesDone.push(candidate);
                    const iceCandidate = JSON.parse(candidate);
                    if (iceCandidate.candidate) {
                        if (MPLAYNET_DEBUG)
                            console.log(`add candidate: ${candidate}`);
                        mesh.addIceCandidate(peer.uuid, iceCandidate);
                    }
                }
            });
        }
    }
}

class LocalSignaling extends BaseSignaling {
    constructor() {
        super();
        this.roomId = '';
        this.username = '';
        this.pairRecordMap = new Map();
        this.timers = [];
    }
    cleanup() {
        this.timers.forEach(timer => {
            if (timer)
                window.clearInterval(timer);
        });
        // HACK: remove timer when where are sure all peers are connected
        // and can safely delete the info
        setTimeout(() => {
            localStorage.removeItem(this.roomId);
            for (const value of this.pairRecordMap.values()) {
                localStorage.removeItem(value);
            }
        }, 5000);
    }
    /* Listen for changes produced by another document */
    listenRoomRecordChanges(roomId, username) {
        onstorage = event => {
            if (event.key === roomId && event.newValue) {
                this.roomRecordChanged(JSON.parse(event.newValue), roomId, username, this._uuid);
            }
        };
    }
    internalHostRoom(roomId, username) {
        this.roomId = roomId;
        this.username = username;
        this.listenRoomRecordChanges(roomId, username);
        return new Promise((resolve) => {
            const roomRecord = new RoomRecord(roomId, []);
            const sRoomRecord = JSON.stringify(roomRecord);
            localStorage.setItem(roomId, sRoomRecord);
            this.roomRecordChanged(roomRecord, roomId, username, this._uuid);
            resolve(true);
        });
    }
    internalJoinRoom(roomId, username) {
        return new Promise((resolve) => {
            const sRoomRecord = localStorage.getItem(roomId);
            if (!sRoomRecord) {
                resolve(false);
            }
            else {
                this.roomId = roomId;
                this.username = username;
                this.listenRoomRecordChanges(roomId, username);
                this.roomRecordChanged(JSON.parse(sRoomRecord), roomId, username, this._uuid);
                resolve(true);
            }
        });
    }
    saveRoomInfo() {
        if (this.roomRecord) {
            localStorage.setItem(this.roomId, JSON.stringify(this.roomRecord));
            this.roomRecordChanged(this.roomRecord, this.roomId, this.username, this._uuid);
        }
    }
    saveIceCandidate(uuid, candidate) {
        const pairRecordName = this.pairRecordMap.get(uuid);
        if (pairRecordName) {
            const sPairingRecord = localStorage.getItem(pairRecordName);
            if (sPairingRecord) {
                const pairingRecord = JSON.parse(sPairingRecord);
                if (uuid === pairingRecord.uuid1) {
                    pairingRecord.iceCandidates2.push(candidate);
                }
                else {
                    pairingRecord.iceCandidates1.push(candidate);
                }
                localStorage.setItem(pairRecordName, JSON.stringify(pairingRecord));
            }
        }
    }
    savePairing(mesh, myIndex, index, peer) {
        if (this.roomRecord) {
            const pairRecordName = `pair_${this.roomRecord.roomId}_${Math.min(index, myIndex)}_${Math.max(index, myIndex)}`;
            this.pairRecordMap.set(peer.uuid, pairRecordName);
            const pairRecord = localStorage.getItem(pairRecordName);
            if (!pairRecord && myIndex < index) {
                const pairingRecord = new PairingRecord(0, '', '', [], [], '', '');
                this.pairingRecordChanged(mesh, pairingRecord, peer, myIndex, index);
            }
            // race condition: if the other peer saves the change before we register the event handler, it doesn't fire.
            /*
            onstorage = event => {
                if (event.key === pairRecordName && event.newValue) {
                    this.pairingRecordChanged(mesh, JSON.parse(event.newValue), peer, myIndex, index);
                }
            }*/
            // Hack: poll the data
            let latestValue = localStorage.getItem(pairRecordName);
            const timer = window.setInterval(() => {
                const currentValue = localStorage.getItem(pairRecordName);
                if (currentValue && currentValue !== latestValue) {
                    latestValue = currentValue;
                    this.pairingRecordChanged(mesh, JSON.parse(currentValue), peer, myIndex, index);
                }
            }, 1000);
            this.timers.push(timer);
        }
    }
    savePairingRecord(info, uuid) {
        const pairRecordName = this.pairRecordMap.get(uuid);
        if (pairRecordName) {
            localStorage.setItem(pairRecordName, JSON.stringify(info));
        }
    }
}

export { LocalSignaling, Mesh, MeshConfig, Message, PeerRecord, getLocalTimestamp, setDebug };
